/* $Id: c18.c,v 1.17 2002/06/09 14:36:35 jym Exp $
 *
 * Coeur du processeur C18
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <unistd.h>
#include "memory.h"
#include "opcode.h"
#include "ansi.h"
#include "spec.h"
#include "display.h"
#include "debug.h"
#include "apm.h"
#include "c18.h"


/* PUBLIC */
extern int proceed;
extern int interrupt_used;
extern int stat_int;



/* PRIVATE */
typedef enum {
	nothing = 0,
	fourth = 0x7,
	three = 0xF8,
	two = 0x1F00,
	one = 0x3E000
} instruction;

#define ZERO(X) X ^= X
#define DROP ZERO(T); dec_top()


/***** REGISTRE / STACK *****/
static unsigned int stack[16];	/* data stack */
static int top;
#define T stack[top]
#define SECOND (top == 0 ? 15 : top-1)
#define S stack[SECOND]

static unsigned int rstack[16];	/* return stack */
static int rtop;
#define R rstack[rtop]

static int B;			/* adress register */
static int A;			/* adress register */
static int P;			/* instruction pointer */


/***** INTERNAL *****/
static int cache;		/* instruction cache */
static instruction prefetch;	/* which instruction is executed in cache (5 bits per instr.) */
static int interruption;
static int P_increment = 1;


/***** Control function *****/
__inline__ static void inc_top(void)
{
	top == 15 ? top = 0 : top++;
}

__inline__ static void dec_top(void)
{
	/* this is also drop */
	stack[top] = -1;
	top == 0 ? top = 15 : top--;
}

__inline__ static void inc_rtop(void)
{
	rtop == 15 ? rtop = 0 : rtop++;
}

__inline__ static void dec_rtop(void)
{
	rstack[rtop] = -1;
	rtop == 0 ? rtop = 15 : rtop--;
}


/* Initialise le CPU
 */
void cpu_init(int start)
{
	int i;

	P = start;
	cache = mem_get(P);
	A = 0;
	B = 0;
	top = 15;
	rtop = 15;
	prefetch = one;
	interruption = 0;
	for(i = 0; i < 16; i++)
		stack[i] = -1;
	for(i = 0; i < 16; i++)
		rstack[i] = -1;
}



/* Remember that fetch pushes the stack, store and binary operations pop it. */
void cpu_op(void)
{
	int instr = 0;

	if(interruption && prefetch == one) {
		inc_rtop();
		R = P;
		P = P & 0x3FF00;
		cache = mem_get(P);
		fflush(stdout);
		prefetch = one;
		interruption = 0;
		return;
	}

	/*
	TODO: faire un log ?
	if(DEBUG_CPU) {
		printf("Instruction (0x%05X:%i): ", P, cpu_get_P(NULL));
	}
	*/

	switch(prefetch) {
		case one:
			prefetch = two;
			instr = (cache >> 13) & 0x1F;
			break;
		case two:
			prefetch = three;
			instr = (cache >> 8) & 0x1F;
			break;
		case three:
			prefetch = fourth;
			instr = (cache >> 3) & 0x1F;
			break;
		case fourth:
			prefetch = nothing;
			instr = (cache & 0x7) << 2;
			break;
		default:
			/* TODO faire un log ?
			printf("*** ERROR: prefetch: 0x%05X\n", prefetch);
			*/
			if(proceed)
				kill(getpid(), SIGINT);
	}
	
	/*
	TODO : faire un log ?
	if(DEBUG_CPU) {
		int tmp = instr;
		printf("[0x%02X] is %s\t", tmp, opcode[tmp]);
		if(tmp >= 2 && tmp <= 7)
			printf("Addr: 0x%05X", cache & 0xFF);
		puts("");
	}
	*/

	temp_cpu_inc();

	switch(instr) {
		
		case 0x00: /* return */
			P = R;
			prefetch = one;
			cache = mem_get(P);
			dec_rtop();
			break;
		
		case 0x02: /* word */
			inc_rtop();
			R = (P + 1) & 0x3FFFF;
			P = (P & 0x3FF00) + (cache & 0xFF);
			prefetch = one;
			cache = mem_get(P);
			break;
		
		case 0x03: /* word ; */
			P = (P & 0x3FF00) + (cache & 0xFF);
			prefetch = one;
			cache = mem_get(P);
			break;
		
		case 0x06: /* if */
			if(T == 0) {
				P = (P & 0x3FF00) + (cache & 0xFF);
				prefetch = one;
				cache = mem_get(P);
			} else {
				prefetch = nothing;
			}
			break;
		
		case 0x07: /* -if */
			if((T&0x20000) == 0) {
				P = (P & 0x3FF00) + (cache & 0xFF);
				prefetch = one;
				cache = mem_get(P);
			} else {
				prefetch = nothing;
			}
			break;

		case 0x08:
			inc_top();
			T = mem_get(P + P_increment);
			P_increment++;
			break;

		case 0x09: /* @+ : fetch from addr in A, increment A */
			inc_top();
			T = mem_get(A);
			A = (A + 1) & 0x3FFFF;
			break;

		case 0x0A: /* @b */
			inc_top();
			T = mem_get(B);
			break;

		case 0x0B: /* @ : fetch from addr in A */
			inc_top();
			T = mem_get(A);
			break;
		
		case 0x0C: /* !r */
			mem_set(R, T);
			R = (R + 1) & 0x3FFFF;
			dec_top();
			break;
		
		case 0x0D: /* !+ : store into addr in A, increment A */
			mem_set(A, T);
			A = (A + 1) & 0x3FFFF;
			dec_top();
			break;
		
		case 0x0E: /* !b */
			mem_set(B, T);
			dec_top();
			break;

		case 0x0F: /* ! : store into addr in A */
			mem_set(A, T);
			dec_top();
			break;
		
		case 0x10: /* +* */
			if(T&0x1) {
				S = ((S + T) & 0x3FFFF) >> 1;
				dec_top();
			}
			break;
		
		case 0x11: /* 2* : shift T left 1 bit */
			T = (T << 1) & 0x3FFFF;
			break;
		
		case 0x12: /* 2/ : shift T right 1 bit (preserve T17) */
			T = ((T & 0x20000) + (T >> 1)) & 0x2FFFF;
			break;
		
		case 0x13: /* - */
			T = (0x3FFFF - T) & 0x3FFFF;
			break;
		
		case 0x14: /* + : Add S to T */
			S = (S + T) & 0x3FFFF;
			dec_top();
			break;
		
		case 0x15: /* or : Exclusive-or S to T */
			S = (S ^ T) & 0x3FFFF;
			dec_top();
			break;
		
		case 0x16: /* and : And S to T */
			S = (S & T) & 0x3FFFF;
			dec_top();
			break;
		
		case 0x17: /* drop : store T nowhere */
			dec_top();
			break;

		case 0x18: /* dup : Duplicate T */
			inc_top();
			T = S;
			break;
		
		case 0x19: /* over : fetch S */
			{
				int tmp = S;
				inc_top();
				T = tmp;
			}
			break;
		
		case 0x1A: /* pop : fetch R */
			inc_top();
			T = R;
			dec_rtop();
			break;
		
		case 0x1B: /* a : fetch A */
			inc_top();
			T = A;
			break;

		case 0x1C: /* nop : do nothing */
			temp_cpu_dec();
			break;

		case 0x1D: /* b! */
			B = T;
			dec_top();
			break;
		
		case 0x1E: /* push : store into R */
			inc_rtop();
			R = T;
			dec_top();
			break;
		
		case 0x1F: /* a! : Store into A */
			A = T;
			dec_top();
			break;
		
		default:
			print_info("Illegal instructions");
			if(proceed)
				kill(getpid(), SIGINT);
	}

	if(prefetch == nothing) {
		P = (P + P_increment) & 0x3FFFF;
		P_increment = 1;
		cache = mem_get(P);
		prefetch = one;
	}
}


/* lorsque l'on a une interruption
 */
void cpu_interrupt(int type)
{
	int c;

	if(!(/*proceed &&*/ (interrupt_used & type)))
		return;

	c = mem_get(REG_CTRL);

	if(c & type) {
		if(proceed)
			kill(getpid(), SIGINT);

		print_info("Interrupt not cleared");
	} else {
		c += type;
	}
	mem_set(REG_CTRL, c);

	interruption = 1;
	stat_int++;
}



/* Some little functions to obtain CPU info
 */
int cpu_get_top(void) { return top; }

char * cpu_get_stack(int i)
{
	char *s = malloc(10);
	bzero(s, 10);
	if(stack[i] == -1)
		memset(s, '-', 7);
	else
		snprintf(s, 10, "0x%05X", stack[i]);

	return s;
}

int cpu_get_rtop(void) { return rtop; }

char * cpu_get_rstack(int i)
{
	char *s = malloc(10);
	bzero(s, 10);
	if(rstack[i] == -1)
		memset(s, '-', 7);
	else
		snprintf(s, 10, "0x%05X", rstack[i]);

	return s;
}

int cpu_get_A(void) { return A; }

int cpu_get_B(void) { return B; }

void cpu_set_P(int i) { P = i; }

int cpu_get_P(int *ip)
{
	if(ip != NULL)
		*ip = P;

	switch(prefetch) {
		case fourth: return 4;
		case three: return 3;
		case two: return 2;
		case one:
		default: return 1;
	}
}
		
