/* $Id: compilo.c,v 1.19 2002/05/10 16:36:00 jym Exp $
 *
 * Assembleur/compilateur forth
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/stat.h>
#include "misc.h"
#include "file.h"
#include "memory.h"
#include "compilo.h"
#include "spec.h"
#include "debug.h"

#define SIZE_MAX_WORD 128

/* faire des enum */
#define SAME_LINE 2
#define END_LINE 40
#define ERROR -1
#define COMMENT 0
#define NEW_WORD 2
#define CALL_WORD 3
#define CALL_VAR 4
#define CODE_JMP 5
#define CODE_VAR 7
#define CODE 6
#define COMPIL 100
#define PREPROC 200

/* data */
extern int interrupt_used;
extern int ip_start;
extern char *opcode[];

static int where = 2;

struct mem_info {
	int line;
	int called;
	int addr;
	char name[SIZE_MAX_WORD];
	struct mem_info *next;
};

struct def_info {
	char name[SIZE_MAX_WORD];
	char replace[SIZE_MAX_WORD];
	struct def_info *next;
};



/* retourne le numero de l'opcode,
 * -1 si pas opcode
 */
static int find_opcode(char *word)
{
	int code = 0;

	while(code < 32) {
		if(cmp(word, opcode[code]))
			return code;
		code++;
	}

	return -1;
}


/* renvoie l'adresse d'un mot/var
 */
static int find_data(struct mem_info *la, char *w)
{
	struct mem_info *ptr = la;

	while(ptr != NULL) {
		if(cmp(ptr->name, w)) {
			ptr->called = 1;
			return ptr->addr;
		} else
			ptr = ptr->next;
	}
	return -1;
}




/* Cherche pour les mots/var non utilise
 */
static void find_data_not_used(struct mem_info *la)
{
	struct mem_info *ptr = la;

	while(ptr != NULL) {
		if(ptr->called == 0)
			printf("%s:%i: warning: %s defined but not used\n", file_get_name(), ptr->line, ptr->name);
		ptr = ptr->next;
	}
}



/* Libere la liste
 */
static void free_data(struct mem_info *la)
{
	struct mem_info *ptr;
	while(la != NULL) {
		ptr = la->next;
		free(la);
		la = ptr;
	}
}


/* Lit un mot sur le fichier
 */
static int read_word(FILE *file, char *word)
{
	char c;
	char *ptr = word;

	while(fread(&c, 1, 1, file) && (ptr - word < SIZE_MAX_WORD)) {
		if(c == ' ' || c == '\t') {
			*ptr = 0;
			return SAME_LINE;
		} else if(c == '\n') {
			*ptr = 0;
			return END_LINE;
		} else {
			*ptr = c;
			ptr++;
		}
	}

	return ERROR;
}



/* On definit un nouveau mot
 */
static int new_word(char *word, int addr, int line, struct mem_info **addr_word)
{
	struct mem_info *ptr = *addr_word;
	struct mem_info *new = jalloc(sizeof(struct mem_info));

	if(DEBUG_COMPIL)
		printf("Line %i: defining %s [0X%05X]\n", line, word, addr);

	strlcpy(new->name, word, SIZE_MAX_WORD);
	new->addr = addr;
	new->line = line;
	new->called = 0;
	new->next = NULL;

	if(ptr != NULL) {
		while(ptr->next != NULL)
			ptr = ptr->next;
		ptr->next = new;
	} else *addr_word = new;

	return 0;
}



/* ecrit le code en memoire. Little complique because
 * c'est 4 opcode par mot de 18 bits.
 */
static int write_opcode(int *addr, int code, int data)
{
	static uint bits = 0x0;
	/* BUG: obliger de retirer where */
	static int increment = 0;

	if(code == 0x8) { /* fetch */
		if(where == -1) {
			bits += 0x2; /* fetch sur slot 4 */
			increment++;
			mem_set((*addr)+increment, data);
			mem_set(*addr, bits);
			(*addr) += increment + 1;
			increment = 0;
			bits = 0;
			where = 2;
		} else {
			bits += code << (where * 5 + 3);
			where--;
			increment++;
			mem_set((*addr)+increment, data);
		}
	} else {
		if(increment != 0) {
		switch(where) {
			case 2:
				break;
			case 1:
				bits += 0x1C00;
			case 0:
				bits += 0xE0;
			case -1:
				bits += 0x7;
				mem_set(*addr, bits);
				bits = 0;
				where = 2;
		}
		(*addr) += increment + 1;
		increment = 0;
		}

	if(code >= 0x02 && code <= 0x07) { /* un call, rcall, return ou if */
		switch(where) {
			case 0:
				bits += 0xE0; /*   . . nop . */
			case -1:
				bits += 0x7;  /*   . . . nop */
				mem_set(*addr, bits);
				(*addr)++;
			case 2:
				bits = 0x38000; /* nop . . . */
		}
		bits += code << 8; /* nop call . . */
		bits += (data & 0xFF); /* nop call addr addr */
		where = -2;
	} else if(code != -1 && code != -2) { /* un code normal */
		if(where == -1) { /* si 4eme position... */
			if((code & 0x3) == 0) { /* si fini par 0x00 */
				bits += (code >> 2) & 0x7;
				where--;
			} else { /* sinon, mets nop et instructions next word */
				bits += 0x7;
				mem_set(*addr, bits);
				(*addr)++;
				bits = code << 13;
				where = 1;
			}
		} else {
			bits += code << (where * 5 + 3);
			where--;
		}
	/* TODO: deux derniers IF peuvent etre simplifie */
	} else if(code == -2) {
		/* donnee immediate doit etre sur un mot */
		if(where != 2) {
			while(where != -1) {
				bits += 0x1C << (where * 5 + 3);
				where--;
			}
			bits += 0x7;
			mem_set(*addr, bits);
			(*addr)++;
		}
		bits = data;
		where = -2;
	} else if(where != 2) { /* utiliser pour completer fin du code */
		switch(where) {
			case 1:
				bits += 0x1C00;
			case 0:
				bits += 0xE0;
			case -1:
				bits += 0x7;
		}
		where = -2;
	}
	} /* end if(code == 8) */

	if(where == -2) {
		mem_set(*addr, bits);
		bits = 0x0;
		(*addr)++;
		where = 2;
	}

	return 0;
}



/* retrouve la directive du preproc %jmp et la remplace 
 * par l'adresse du mot en cours. Si plusieurs ?
 */
int find_adr(struct mem_info **first, char *word, int addr)
{
	struct mem_info *ptr = *first;
 	uint bits;

	while(ptr != NULL) {
		if(cmp(ptr->name, word)) {
			ptr->called = 1;
			bits = 0x38300 + (addr & 0xFF);
			mem_set(ptr->addr, bits);
		}
		ptr = ptr->next;
	}

	return 0;
}




/* Retourner erreur quand commande inconnu
 */
static int preprocesseur(FILE *file, char *first, struct mem_info **var, char *start, int state, int *addr, int line, struct def_info **define, struct mem_info **adr)
{
	int rv = SAME_LINE;
	char *word = jalloc(SIZE_MAX_WORD);
	char *size = jalloc(SIZE_MAX_WORD);

	DENTER;

	if(state == END_LINE) {
	} else if(cmp(first, "%start")) {
		read_word(file, start);
		return 0;
	} else if(cmp(first, "%data")) {
		read_word(file, word);
		new_word(word, *addr, line, var);
		read_word(file, size);
		/* verifier si c bien la qu'on le met */
		find_adr(adr, word, *addr);
		/* on reserve l'espace pour la variable (en se calant sur 
		 * le prochain mot automatiquement) */
		write_opcode(addr, -2, (int)strtol(size, (char **)NULL, 0));
	} else if(cmp(first, "%bss")) {
		int tmp;
		read_word(file, word);
		new_word(word, *addr, line, var);
		read_word(file, size);
		/* verifier si c bien la qu'on le met */
		find_adr(adr, word, *addr);
		/* cette fois on reserve size mots */
		tmp = (int)strtol(size, (char **)NULL, 0);
		while(tmp--) {
			write_opcode(addr, -2, 0);
		}
	} else if(cmp(first, "%define")) {
		struct def_info *new_def = jalloc(sizeof(struct def_info));
		struct def_info *tmp = *define;
		read_word(file, word);
		strlcpy(new_def->name, word, SIZE_MAX_WORD);
		read_word(file, word);
		strlcpy(new_def->replace, word, SIZE_MAX_WORD);
		if(!*define)
			*define = new_def;
		else {
			while(tmp->next)
				tmp = tmp->next;
			tmp->next = new_def;
		}
	} else if(cmp(first, "%jmp")) {
		struct mem_info *new_adr = jalloc(sizeof(struct mem_info));
		struct mem_info *tmp = *adr;
		read_word(file, word);
		strlcpy(new_adr->name, word, SIZE_MAX_WORD);
		new_adr->line = line;
		new_adr->addr = *addr;
		if(!*adr) {
			*adr = new_adr;
		} else {
			while(tmp->next)
				tmp = tmp->next;
			tmp->next = new_adr;
		}
		/* et on reserve un espace */
		write_opcode(addr, -2, 0);
	} else if(cmp(first, "%interrupt")) {
		read_word(file, word);
		if(ishexa(word)) interrupt_used = (int)strtol(word, (char **)NULL, 0);
		else printf("%s:%i: warning: bad interrupt number\n", file_get_name(), line);
	} else if(cmp(first, "%cpu")) {
		read_word(file, word);
		printf("%s:%i: warning: multiple CPU not implemented!\n", file_get_name(), line);
	} else if(first[0] == '#') {
		while(rv != END_LINE) {
			rv = read_word(file, word);
			bzero(word, SIZE_MAX_WORD);
		}
	} else {
		free(word);
		DLEAVE;
		return -1;
	}

	free(word);
	DLEAVE;

	return 0;
}




/* Pour trouver le point de depart
 */
static int find_start_point(struct mem_info *mem, char *point)
{
	int addr = 0;

	if(point && strlen(point) > 0) {
		if((addr = find_data(mem, point)) != -1) {
			return addr;
		} else {
			printf("%s: warning: can't find start point '%s'\n", file_get_name(), point);
			return DEFAULT_IP;
		}
	}

	printf("%s: warning: start point not defined\n", file_get_name());
	return DEFAULT_IP;
}






/* Lorsque c'est un saut. Il faut differencier les 2 types de saut.
 */
static int code_call_jmp(FILE *file, int addr, int *current_addr)
{
	int rv = 0;
	char next[SIZE_MAX_WORD];
	long whereami;

	whereami = ftell(file);

	bzero(next, SIZE_MAX_WORD);

	while(strlen(next) == 0) {
		rv = read_word(file, next);
	}

	if(*next == ';') {
		write_opcode(current_addr, 0x3, addr);
	} else {
		write_opcode(current_addr, 0x2, addr);
		fseek(file, whereami, SEEK_SET);
		rv = SAME_LINE;
	}

	/* TODO: pas top */
	return rv;
}


/* Pour rechercher une macro (%define)
 */
int find_define(struct def_info *def, char **w)
{
	while(def) {
		if(cmp(def->name, *w)) {
			strlcpy(*w, def->replace, SIZE_MAX_WORD);
		}
		def = def->next;
	}
	return 0;
}

/* Parce que pour les if, faut trouver l'adresse du 'then'
 */
static int new_if(int addr, int line, struct mem_info **list_if)
{
	struct mem_info *ptr = *list_if;
	struct mem_info *new = jalloc(sizeof(struct mem_info));

	new->line = line;
	new->addr = addr;
	new->next = NULL;

	if(ptr != NULL) {
		while(ptr->next != NULL) {
			ptr = ptr->next;
		}
		ptr->next = new;
	} else *list_if = new;

	return 0;
}





/* Retrouve le dernier if et mets l'adresse sur les 8 derniers bits
 */
static int find_last_if(int addr, int line, struct mem_info **list_if)
{
	int number = 0;
	uint bits;
	struct mem_info *last = *list_if;
	struct mem_info *ptr = *list_if;

	if(!ptr) {
		printf("%s:%i: error: then without if\n", file_get_name(), line);
		return -1;
	}

	while(ptr->next != NULL) {
		last = ptr;
		ptr = ptr->next;
		number++;
	}

	if(ptr->addr & 0xFFFFFF00) {
		printf("%s:%i: error: jump too big (from line %i)\n", file_get_name(), line, ptr->line);
		return -1;
	}
	bits = mem_get(ptr->addr);
	bits = (bits & 0x3FF00) + (addr & 0xFF);
	mem_set(ptr->addr, bits);

	free(ptr);

	if(number == 0) {
		*list_if = NULL;
	} else {
		last->next = NULL;
	}

	return 0;
}



/* On compile le fichier
 */
int compile(int fd, int start)
{
	int current_addr = start;
	int current_line = 1;
	int rv;
	int code;
	int last = END_LINE;
	char *word = jalloc(SIZE_MAX_WORD);
	char *start_point = jalloc(SIZE_MAX_WORD);
	FILE *source = NULL;
	struct mem_info *first_word = NULL;
	struct mem_info *first_variable = NULL;
	struct mem_info *first_if = NULL;
	struct mem_info *first_adr = NULL;
	struct def_info *definition = NULL;

	source = fdopen(fd, "r");

	if(DEBUG_COMPIL)
		printf("Assembling at address: 0x%05X\n", start);

	while(!feof(source)) {
		bzero(word, SIZE_MAX_WORD);
		rv = read_word(source, word);
		if(rv == ERROR) break;
		find_define(definition, &word);
		if(strlen(word) == 0) {
		} else if((word[0] == '%' || word[0] == '#')/* && last == END_LINE*/) {
			
			if(preprocesseur(source, word, &first_variable, start_point, rv, &current_addr, current_line, &definition, &first_adr))
				printf("%s:%i: error: unknow preproc command\n",
					file_get_name(), current_line);
			rv = END_LINE;
			
		} else if((code = find_opcode(word)) != -1) {
			
			if(code == 0x6 || code == 0x7) {
				/* this is a little BUG pour ? */
				if(where == 0)
					new_if(current_addr+1, current_line, &first_if);
				else
					new_if(current_addr, current_line, &first_if);
			}
			write_opcode(&current_addr, code, 0);

		} else if((code = find_data(first_word, word)) != -1) {
			
			/* TODO: Et si commentaires avant prochain mot ? */
			rv = code_call_jmp(source, code, &current_addr);

		} else if((code = find_data(first_variable, word)) != -1) {

			write_opcode(&current_addr, 0x08, code);

		} else if(ishexa(word)) {

			write_opcode(&current_addr, 0x08, (int)strtol(word, (char **)NULL, 0));

		} else {
			write_opcode(&current_addr, -1, 0);
			if(cmp(word, "then")) find_last_if(current_addr, current_line, &first_if);
			else new_word(word, current_addr, current_line, &first_word);

		}
		/* si on attendait ce mot pour avoir l'adresse */
		find_adr(&first_adr, word, current_addr);
		
		if(rv == END_LINE) {
			current_line++;
		}

		last = rv;
	}

	write_opcode(&current_addr, -1, 0);

	ip_start = find_start_point(first_word, start_point);

	find_data_not_used(first_word);
	find_data_not_used(first_variable);
	find_data_not_used(first_adr);

	free_data(first_word);
	free_data(first_variable);
	free_data(first_adr);

	free(word);
	free(start_point);

	return 0;
}

