/* $Id: display.c,v 1.2 2002/06/09 14:36:35 jym Exp $
 *
 * petit fonction d'affichage (ncurses)
 */

#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <curses.h>
#include <string.h>
#include "display.h"
#include "c18.h"
#include "misc.h"
#include "memory.h"
#include "ansi.h"
#include "spec.h"


extern char *opcode[];
extern int interrupt_used;
extern double mips;
extern double idle;

void update_register(int reg, int data)
{
	switch(reg) {
		case REG_A:
		case REG_B:
		case REG_C:
	}
}



void dump_memory(void)
{
	int i = 0;
	int ip;
	int which = 0;
	static int current_ip;
	int tmp;
	int first;
	int second;
	int thirteen;
	int fourtheen;
	int increment = 0;

	which = cpu_get_P(&ip);

	if(ip < current_ip || ip >= (current_ip+16)) {
		current_ip = ip;
	}

	while(i < 16) {
		tmp = mem_get(current_ip + i);

		move(8 + i, 34);

		printw("0x%05X: 0x%05X  ", current_ip + i, tmp);

		if(increment > 0) {
			printw("0x%05X", tmp);
			increment--;
			i++;
			continue;
		}


		first = (tmp&0x3E000) >> 13;
		second = (tmp&0x1F00) >> 8;
		thirteen = (tmp&0xF8) >> 3;
		fourtheen = (tmp&0x7) << 2;

		if(which == 1 && current_ip+i == ip) { attrset(A_BOLD); printw("%5s  ", opcode[first]); attrset(A_NORMAL); }
		else printw("%5s  ", opcode[first]);
		if(first == 8) increment++;

		if(which == 2 && current_ip+i == ip) { attrset(A_BOLD); printw("%5s  ", opcode[second]); attrset(A_NORMAL); }
		else printw("%5s  ", opcode[second]);
		if(second == 8) increment++;

		if(second >= 2 && second <= 7) {
			printw("0x%02X", tmp & 0xFF);
		} else {
			if(which == 3 && current_ip+i == ip) { attrset(A_BOLD); printw("%5s  ", opcode[thirteen]); attrset(A_NORMAL); }
			else printw("%5s  ", opcode[thirteen]);
			if(thirteen == 8) increment++;

			if(which == 4 && current_ip+i == ip) { attrset(A_BOLD); printw("%5s  ", opcode[fourtheen]); attrset(A_NORMAL); }
			else printw("%5s  ", opcode[fourtheen]);
			if(fourtheen == 8) increment++;
		}
		i++;
	}
}



void display_update(void)
{
	int i = 0;
	int tmp;

	erase();
	move(0, 0);
	cpu_get_P(&i);
	printw("P:\t0x%05X\t\t\t\t|", i);
	move(1, 0);
	printw("A:\t0x%05X\t\t\t\t|", cpu_get_A());
	move(2, 0);
	printw("B:\t0x%05X\t\t\t\t|", cpu_get_B());
	move(3, 0);
	tmp = mem_get(REG_CTRL);
	printw("C:\t0x%05X (", tmp);
	for(i = 17; i >= 0; i--)
		printw("%1i", (tmp >> i) & 1);
	printw(")\t|");
	move(4, 0);
	printw("Mips:\t%04.02f\t\t\t\t|", mips);
	move(5, 0);
	printw("Int:\t");
	if(interrupt_used & INTERRUPT(C_ISERIAL)) printw("serial ");
	if(interrupt_used & INTERRUPT(C_ITIMER)) printw("timer ");
	if(interrupt_used & INTERRUPT(C_IKBD)) printw("kbd ");
	if(interrupt_used & INTERRUPT(C_IAUDIO)) printw("audio ");
	if(interrupt_used & INTERRUPT(C_IRADIO)) printw("radio ");
	move(5, 40);
	printw("|");
	move(6,0);
	printw("Idle:\t%04.02f%%\t\t\t\t|", idle);
	move(7, 0);
	for(i = 0; i < 40; i++)
		printw("_");
	printw("|");
	for(i = 41; i < 80; i++)
		printw("_");
	for(i = 0; i < 16; i++) {
		move(i+8, 0);
		printw("S%2i:  %s   R%2i:  %s  |", 15-i, cpu_get_stack(15-i), 15-i, cpu_get_rstack(15-i));
	}

	dump_memory();

	move(0, 42);

	refresh();
}



void display_init(void)
{
	initscr();
	start_color();
	curs_set(0);
	display_update();
}


void update_stack(int level, int data)
{
	move(level+5, 7);
	printw("0x%05X", data);
}


void update_rstack(int level, int data)
{
	move(level+5, 23);
	printw("0x%05X", data);
}

/* data = P et instr = quel slot */
void update_P(int data, int instr)
{
}


void print_info(char *str, ...)
{
	char *buf = jalloc(128);
	va_list ap;
	WINDOW *win;

	va_start(ap, str);
	vsnprintf(buf, 128, str, ap);
	va_end(ap);

	win = newwin(6, strlen(buf)+1, 9, (80-strlen(buf))/2);
	wmove(win, 1, strlen(buf)/2-6);
	wprintw(win, "*** ERROR ***");
	wmove(win, 3, 1);
	wprintw(win, buf);
	box(win, '|', '-');
	wrefresh(win);

	getch();
	delwin(win);
	
	display_update();
	free(buf);
}
