/* $Id: file.c,v 1.7 2002/02/05 23:38:30 jym Exp $
 *
 * Gestion lecture/ecriture du debuggueur
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "memory.h"
#include "debug.h"
#include "misc.h"
#include "ansi.h"
#include "compilo.h"


int is_code_loaded = 0;

static int destination;
static int data_where = 0;	/* 18bits -> 16bits */
static int memory_cache = 0;	/* cache for 18 -> 16 */
static int fd = 0;
#define PATHSIZE 128
static char fichier[PATHSIZE] = { "nothing" };
static u_char code_cache = 0;
static enum { un = 1, deux = 2, trois = 3, quatre = 4 } where = un;


/* ferme le fichier
 */
int file_close(void)
{
	if(fd == 0)
		return -1;

	close(fd);
	fd = 0;
	is_code_loaded = 0;
	strlcpy(fichier, "nothing", PATHSIZE);
	return 0;
}

/* Ouvre le fichier
 */
int file_open(char *s)
{
	file_close();

	printf("Openning %s ", s);

	if((fd = open(s, O_RDONLY)) < 0) {
		puts(PRINT_FAILED);
		//perror("open");
		fd = 0;
		return -1;
	}

	puts(PRINT_OK);
	strlcpy(fichier, s, PATHSIZE);
	is_code_loaded = 1;

	return 0;
}

/* Permet de savoir quel fichier est utilise
 */
char *file_get_name(void)
{
	return fichier;
}

/* Charge un opcode.
 * On charge 8 bits par 8 bits.
 * On retourne quand l'instruction est complete
 * L'instrucion est retourne sur un int
 */
static uint file_opcode(void)
{
	uint code = -1;
	u_char buf = 0;

	switch(where) {
		case un:
			if(read(fd, &buf, sizeof buf) <= 0) return -1;
			code = buf << 10;
			if(read(fd, &buf, sizeof buf) <= 0) return -1;
			code += buf << 2;
			if(read(fd, &buf, sizeof buf) <= 0) return -1;
			code += buf >> 6;
			code_cache = buf & 0x3F;
			where = deux;
			break;
		case deux:
			code = code_cache << 12;
			if(read(fd, &buf, sizeof buf) <= 0) return -1;
			code += buf << 4;
			if(read(fd, &buf, sizeof buf) <= 0) return -1;
			code += buf >> 4;
			code_cache = buf & 0x0F;
			where = trois;
			break;
		case trois:
			code = code_cache << 14;
			if(read(fd, &buf, sizeof buf) <= 0) return -1;
			code += buf << 6;
			if(read(fd, &buf, sizeof buf) <= 0) return -1;
			code += buf >> 2;
			code_cache = buf & 0x03;
			where = quatre;
			break;
		case quatre:
			code = code_cache << 16;
			if(read(fd, &buf, sizeof buf) <= 0) return -1;
			code += buf << 8;
			if(read(fd, &buf, sizeof buf) <= 0) return -1;
			code += buf;
			code_cache = 0;
			where = un;
	}

	return code;
}

/* Charge le contenu d'un fichier en memoire
 */
int file_load_binary(int start)
{
	int ptr = start;
	uint opcode;

	printf("Loading binary file '%s' [0x%08X]", fichier, start);

	if(fd == 0) {
		puts(PRINT_FAILED);
		return -1;
	}

	code_cache = 0;
	where = un;

	lseek(fd, 0, SEEK_SET);
	while((opcode = file_opcode()) != -1) {
		mem_set(ptr++, opcode);
		if(DEBUG_LOAD)
			printf(".");
	}

	puts(PRINT_OK);

	mem_lock();

	return 0;
}


/* on charge en compilant un fichier source
 */
int file_load_source(int start)
{

	if(fd)
		if(!compile(fd, start)) {
			printf("Compiling file '%s' [0x%05X]", fichier, start);
			puts(PRINT_OK);
			mem_lock();
			return 0;
		}

	printf("Compiling file '%s' [0x%05X]", fichier, start);
	puts(PRINT_FAILED);

	return -1;
}




/* ecrit un octet dans le fichier
 */
static int write_byte(u_char *b)
{
	if(DEBUG_LOAD)
		printf("write to disk: 0x%02X\n", *b);
	if(write(destination, b, 1) <= 0)
		perror("write");

	return 0;
}



/* Ecrit 16 bits et sauvegarde les 2 bits non-ecrits
 */
static int write_instr(int instr)
{
	u_char b = 0;

	if(DEBUG_LOAD) {
		bin_mem(instr);
	}

	if(data_where != 0) {
		b = memory_cache << (8-data_where);
	} else {
		data_where = 0;
	}
		
	b += instr >> (10 + data_where);
	data_where += 2;
	write_byte(&b);

	b = instr >> data_where;
	write_byte(&b);

	if(data_where == 8) {
		b = instr;
		write_byte(&b);
		data_where = 0;
	}

	memory_cache = instr;

	return 0;
}


void file_save_binary(char *file, int start, int end)
{
	uint instr = 0x0;
	int current_addr = start;

	if((destination = open(file, O_RDWR|O_CREAT|O_TRUNC, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH)) < 0) {
		perror("Can't open file");
		return;
	}

	printf("Save memory dump [0x%08X -> 0x%08X] into %s\n", start, end, file);

	while(current_addr < end) {
		instr = mem_get(current_addr);
		write_instr(instr);
		current_addr++;
	}

	while(data_where != 0) {
		instr = 0;
		write_instr(instr);
	}

	close(destination);
}

/* EOF */
