/* $Id: memory.c,v 1.21 2002/05/14 15:51:25 jym Exp $
 *
 * Gestion memoire
 *
 * La memoire interne au CPU sera apres la RAM
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include "spec.h"
#include "c18.h"
#include "memory.h"
#include "display.h"
#include "misc.h"
#include "video.h"
#include "kbd.h"
#include "apm.h"
#include "ansi.h"
#include "system.h"
#include "debug.h"


extern char *opcode[];
extern int proceed;

static int *mem = NULL;
static int mem_size = 192;
static int rom_lock = 0;

/* Pour allouer la memoire
 */
int mem_init(void)
{
	printf("Memory (%i word)", mem_size);

	mem_free();

	mem = malloc (mem_size * sizeof(int));

	if(!mem) {
		//puts(PRINT_FAILED);
		quit(0);
	}

	bzero (mem, mem_size * sizeof(int));

	//puts (PRINT_OK);

	return 0;
}



/* Entree
 */
uint in(int addr)
{
	switch(addr) {
		case REG_KBD:
		case REG_KBD+1:
			return kbd_get(addr - REG_KBD);
		case REG_TIMER:
			return timer_get();
		case REG_SENSOR:
			return sensor_get();
		default:
	}
	if(addr >= mem_size || addr < 0) {
		print_info("Segfault (reading) address: 0x%05X\n", addr);
		if(proceed)
			kill(getpid(), SIGINT);
		return -1;
	}

	return mem[addr] & 0x3FFFF;
}



/* Sortie
 */
int out(int addr, uint data)
{
	switch(addr) {
		case REG_KBD:
		case REG_KBD+1:
			return kbd_set(addr - REG_KBD);
		case REG_TIMER:
		case REG_SENSOR:
			//puts("Read-only address");
			return 0;
		case REG_VIDADDR:
			if(mem[REG_VIDDATA] & 0x30000) fb_set(data, in(REG_VIDDATA) & 0xFFFF);
			else mem[REG_VIDDATA] = fb_get(data);
		default:
	}
	if(addr >= mem_size || addr < 0) {
		print_info("Segfault (writing) address: 0x%05X\n", addr);
		if(proceed)
			kill(getpid(), SIGINT);
		return -1;
	}

	mem[addr] = data & 0x3FFFF;

	return 0;
}



uint mem_get(int addr)
{
	return in(addr);
}

int mem_set(int addr, uint data)
{
	return out(addr, data);
}


int mem_get_size(void)
{
	return mem_size;
}


/* Pour afficher en binaire le contenu (18 bits)
 * d'un emplacement memoire
 */
void bin_mem(int d)
{
	int i = 17;
	while(i >= 0) {
		printf("%i", (d >> i) & 1);
		i--;
	}
	printf("b");
}



/* Pour liberer la memoire
 */
void mem_free(void)
{
	if(mem)
		free(mem);
	mem = NULL;
}

/* Pour debloquer la ROM
 */
void mem_unlock(void)
{
	rom_lock = 0;
}

/* Pour bloquer la ROM
 */
void mem_lock(void)
{
	rom_lock = 1;
}
