/* $Id: mon.c,v 1.22 2002/06/09 14:36:35 jym Exp $
 *
 */
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <curses.h>
#include "file.h"
#include "misc.h"
#include "c18.h"
#include "spec.h"
#include "memory.h"
#include "system.h"
#include "video.h"
#include "kbd.h"
#include "display.h"



uint verbose = 0;
int ip_start = DEFAULT_IP;
int load_start = DEFAULT_LOAD;
int proceed = 0;
int num_of_cpu = 1;
int interrupt_used = 0;
int stat_int = 0;
int stat_radior = 0;
int stat_radiow = 0;
int stat_audior = 0;
int stat_audiow = 0;
int stat_videor = 0;
int stat_videow = 0;
int stat_gps = 0;
int stat_apm = 0;
int stat_sensor = 0;
int stat_kbd = 0;
int stat_timer = 0;

extern int is_code_loaded;
static double num_of_instr = 0;
double idle_instruction = 0;
double idle = 0;
double mips = 0;
static struct rusage *ruse = NULL;


static void stop_process(int s)
{
	proceed = 0;
	timer_stop();
	signal(SIGINT, quit);
	move(4, 8);
	mips = (double)(num_of_instr / end_clock(ruse)) / 1000000.0;
	if(num_of_instr) idle = idle_instruction * 100 / num_of_instr;
	printw("%04.02f%%", mips);
	move(0, 42);
	printw("Stop.");
	refresh();
}


/* prochaine instructions
 */
void next(void)
{
	cpu_op();
	display_update();
}


/* lance le prog
 */
void run(void)
{
	signal(SIGINT, stop_process);
	move(0, 42);
	printw("Running...");
	refresh();
	proceed = 1;
	num_of_instr = 0;
	idle_instruction = 0;
	ruse = start_clock();
	timer_init();                     
	for(;proceed;) {
		cpu_op();
		num_of_instr++;
	}
}


/* modifier un parametre (debug, ...)
 */
void set(void)
{
}



/* afficher statistique
 */
void status(void)
{
	WINDOW *win;

	win = newwin(13, 25, 7, 27);
	box(win, '|', '-');
	wmove(win, 1, 2);
	wprintw(win, "-= Statistique IO =-");
	wmove(win, 3, 1);
	wprintw(win, "interrupt:\t%i", stat_int);
	wmove(win, 4, 1);
	wprintw(win, "video:\t\t%i/%i", stat_videor, stat_videow);
	wmove(win, 5, 1);
	wprintw(win, "audio:\t\t%i/%i", stat_audior, stat_audiow);
	wmove(win, 6, 1);
	wprintw(win, "radio:\t\t%i/%i", stat_radior, stat_radiow);
	wmove(win, 7, 1);
	wprintw(win, "apm:\t\t%i", stat_apm + stat_sensor);
	wmove(win, 8, 1);
	wprintw(win, "kbd:\t\t%i", stat_kbd);
	wmove(win, 9, 1);
	wprintw(win, "timer:\t\t%i", stat_timer);

	wrefresh(win);
	getch();
	delwin(win);
	display_update();
}


/* afficher l'aide
 */
void help(void)
{
	WINDOW *win;

	win = newwin(13, 25, 7, 27);
	box(win, '|', '-');
	wmove(win, 1, 7);
	wprintw(win, "-= Help =-");
	wmove(win, 3, 2);
	wprintw(win, "d : show memory dump");
	wmove(win, 4, 2);
	wprintw(win, "h : show help");
	wmove(win, 5, 2);
	wprintw(win, "r : run code");
	wmove(win, 6, 2);
	wprintw(win, "space : next instr");
	wmove(win, 7, 2);
	wprintw(win, "s : set variable");
	wmove(win, 8, 2);
	wprintw(win, "S : show stat");
	wmove(win, 9, 2);
	wprintw(win, "q : quit");
	wrefresh(win);
	getch();
	delwin(win);
	display_update();
}



/* afficher un autre emplacement memoire (nelle fenetre)
 */
void dump(void)
{
	char *buf = jalloc(20);
	WINDOW *win;

	win = newwin(6, 24, 7, 27);
	box(win, '|', '-');
	wmove(win, 1, 3);
	wprintw(win, "-= dump memory =-");
	wmove(win, 3, 1);
	wprintw(win, "From : ");
	wrefresh(win);
	wgetnstr(win, buf, 20);
	delwin(win);
	display_update();
}


/* affiche ou pas (ou desactive refresh) video
 */
void video(void)
{
}


/* envoie une interruption
 */
void interrupt(void)
{
	char c;
	WINDOW *win;

	win = newwin(6, 24, 7, 27);
	box(win, '|', '-');
	wmove(win, 1, 3);
	wprintw(win, "-= Send interrupt =-");
	wmove(win, 3, 2);
	wprintw(win, "keypad for keyboard");
	wmove(win, 4, 2);
	wprintw(win, "space for timer");
	wrefresh(win);
	c = getch();
	delwin(win);
	display_update();
	/* TODO envoyer interruption */
#ifdef HAVE_SDL
	if(c == ' ')
		timer_update(30, 0);
#else
	if(c == ' ')
		timer_update(0);
#endif
}



/*
 */
void option(void)
{

}



int main(int argc, char *argv[])
{
	char c = 0;

	mem_init();
	fb_init();
	kbd_init();

	if(argc > 2) {
		printf("Usage: %s [file]\n\n", argv[0]);
		return 0;
	} else if(argc == 2) {
		if(file_open(argv[1]))
			quit(0);
		file_load_source(DEFAULT_LOAD);
	}

	cpu_init(ip_start);
	display_init();

	signal(SIGINT, quit);
	signal(SIGTERM, quit);

	while(c != 'q') {
		c = getch();
		switch(c) {
			case 'q': quit(0); break;
			case 'd': dump(); break;
			case 'r': run(); break;
			case 's': set(); break;
			case 'S': status(); break;
			case 'h': help(); break;
			case ' ': next(); break;
			case 'v': video(); break;
			case 'i': interrupt(); break;
			case 'o': option(); break;
			default:
				printw("?");
		}
		move(0, 42);
		display_update();
	}

	quit(0);

	return 0;
}
