/* $Id: mon.c,v 1.20 2002/04/24 11:31:41 jym Exp $
 *
 * Debuggueur
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <termios.h>
#include <signal.h>
#include <fcntl.h>
#include <ctype.h>
#include <time.h>
#include "memory.h"
#include "video.h"
#include "misc.h"
#include "file.h"
#include "system.h"
#include "spec.h"
#include "kbd.h"
#include "c18.h"
#include "debug.h"
#include "ansi.h"

#define NUM_OF_ARG 4

/* Variable globale interne */
uint verbose = 0;
int ip_start = DEFAULT_IP;
int load_start = DEFAULT_LOAD;
int proceed = 0;
int num_of_cpu = 1;
int interrupt_used = 0; //INTERRUPT(C_IKBD) | INTERRUPT(C_IAUDIO) | INTERRUPT(C_IRADIO) | INTERRUPT(C_ITIMER) | INTERRUPT(C_ISERIAL);
int stat_int = 0;
int stat_radior = 0;
int stat_radiow = 0;
int stat_audior = 0;
int stat_audiow = 0;
int stat_videor = 0;
int stat_videow = 0;
int stat_gps = 0;
int stat_apm = 0;
int stat_sensor = 0;
int stat_kbd = 0;
int stat_timer = 0;
struct termios cons_state;

extern int is_code_loaded;
static double num_of_instr = 0;
static char *cmd_line[NUM_OF_ARG] = { NULL, NULL, NULL, NULL };
static struct rusage *ruse = NULL;




/* Ctrl-C lorsque cpu fonctionne
 */
static void stop_process(int s)
{
	proceed = 0;
	timer_stop();
	fflush(stdout);
	usleep(1000);
	printf("\nExecution stopped after %.0f instructions. %.2f MIPS\n", num_of_instr, (double)(num_of_instr / end_clock(ruse)) / 1000000.0);
	signal(SIGINT, quit);
}


/* dump de la memoire
 */
static int dump(void)
{
	int x, y = -1;

	if(cmd_line[1]) {
		x = (int)strtol(cmd_line[1], (char **)NULL, 0);
	} else {
		cpu_get_P(&x);
		x -= 5;
	}

	if(x < 0) x = 0;

	if(cmd_line[2]) {
		y = (int)strtol(cmd_line[2], (char **)NULL, 0);
	} else {
		y = x + 10;
	}

	if(y < x) y = x + 10;

	mem_dump(x, y);

	return 0;
}





/* juste pour afficher les interruptions utilise
 */
void int_init(void)
{
	printf("interrupt  :\t");
	if(interrupt_used & INTERRUPT(C_IKBD))
		printf("keyboard ");
	if(interrupt_used & INTERRUPT(C_IRADIO))
		printf("radio ");
	if(interrupt_used & INTERRUPT(C_IAUDIO))
		printf("audio ");
	if(interrupt_used & INTERRUPT(C_ITIMER))
		printf("timer ");
	if(interrupt_used & INTERRUPT(C_ISERIAL))
		printf("serial ");
	puts("");
}

void debug_init(void)
{
	printf("debug      :\t");
	if(DEBUG_CPU) printf("cpu ");
	if(DEBUG_MEM) printf("mem ");
	if(DEBUG_COMPIL) printf("compil ");
	if(DEBUG_VIDEO) printf("video ");
	if(DEBUG_KBD) printf("kbd ");
	if(DEBUG_TIMER) printf("timer ");
	if(DEBUG_LOAD) printf("load ");
	if(DEBUG_INT) printf("int ");
	if(DEBUG_SENSOR) printf("sensor ");
	puts("");
}


/* Affichage de quelques statistique sur la VM
 * (nombre de IO, nombre d'interruption...)
 */
int status(void)
{
	printf( "interrupt  :\t%i\n"
		"video r/w  :\t%i/%i\n"
		"audio r/w  :\t%i/%i\n"
		"radio r/w  :\t%i/%i\n"
		"gps read   :\t%i\n"
		"apm read   :\t%i\n"
		"sensor read:\t%i\n"
		"kbd read   :\t%i\n"
		"timer read :\t%i\n"
		, stat_int, stat_videor, stat_videow, stat_audior, stat_audiow, stat_radior, stat_radiow, stat_gps, stat_apm, stat_sensor, stat_kbd, stat_timer);
	return 0;
}

/* Modification des variables internes
 */
static int set(void)
{
	if(cmd_line[1]) {
		if(!strncasecmp(cmd_line[1], "file", 4)) {
			if(cmd_line[2]) {
				file_open(cmd_line[2]);
			}
		} else if(!strncasecmp(cmd_line[1], "interrupt", 9)) {
			if(cmd_line[2]) {
				interrupt_used = (int)strtol(cmd_line[2], (char **)NULL, 0);
				int_init();
			}
		} else if((!strncasecmp(cmd_line[1], "verbose", 7)) || (!strncasecmp(cmd_line[1], "debug", 5))) {
			if(cmd_line[2]) {
				verbose = (int)strtol(cmd_line[2], (char **)NULL, 0);
				debug_init();
			}
		} else if(!strncasecmp(cmd_line[1], "start", 5)) {
			if(cmd_line[2]) {
				ip_start = (int)strtol(cmd_line[2], (char **)NULL, 0);
				printf("Setting load_start to %i\n", ip_start);
			}
		} else if(!strncasecmp(cmd_line[1], "P", 5)) {
			if(cmd_line[2]) {
				ip_start = (int)strtol(cmd_line[2], (char **)NULL, 0);
				cpu_set_P(ip_start);
				printf("Setting P to %i\n", ip_start);
			}
		} else 
			puts("Unknow var");
	} else {
		printf( "file       :\t%s\n"
			"load at    :\t0x%05X\n"
			"cpu start  :\t0x%05X\n"
			"num of cpu :\t%i\n"
			"memory     :\t0x%05X -> 0x%05X\n"
			"IO enable  :\tkeyboard radio audio video gps timer apm\n"
			, file_get_name(), load_start, ip_start, num_of_cpu, IRAM_START, IRAM_END);
		int_init();
		debug_init();
	}

	return 0;
}


/* Chargement d'un code binaire
 */
static void load(void)
{
	if(cmd_line[1]) {
		if(ishexa(cmd_line[1]))
			file_load_binary((int)strtol(cmd_line[1], (char **)NULL, 0));
		else
			puts("error: Bad addr.");
	} else {
		file_load_binary(load_start);
	}

}



/* Compilation/assemblage d'un code source
 */
static void comp(void)
{
	if(cmd_line[1]) {
		if(ishexa(cmd_line[1]))
			file_load_source((int)strtol(cmd_line[1], (char **)NULL, 0));
		else
			puts("error: bad addr.");
	} else {
		file_load_source(load_start);
		int_init();
	}
}



/* Sauvegarde d'une partie de la memoire dans un fichier
 */
static void save(void)
{
	if(cmd_line[1]) {
		if(cmd_line[2]) {
			if(ishexa(cmd_line[2])) {
				if(cmd_line[3]) {
					if(ishexa(cmd_line[3])) {
						file_save_binary(cmd_line[1], (int)strtol(cmd_line[2], (char **)NULL, 0), (int)strtol(cmd_line[3], (char **)NULL, 0));
					} else {
						puts("error: bad addr for end.");
					}
				} else {
					file_save_binary(cmd_line[1], (int)strtol(cmd_line[2], (char **)NULL, 0), mem_get_size());
				}
			} else {
				puts("error: bad addr for start.");
			}
		} else {
			file_save_binary(cmd_line[1], 0, mem_get_size());
		}
	} else {
		file_save_binary(DEFAULT_SAVE_NAME, 0, mem_get_size());
	}
}




/* Initialisation des differents systemes
 */
static void init(void)
{
	if(cmd_line[1]) {
		if(!strncasecmp(cmd_line[1], "cpu", strlen(cmd_line[1]))) {
			cpu_init(ip_start);
		} else if(!strncasecmp(cmd_line[1], "fb", strlen(cmd_line[1]))) {
			fb_init();
		} else if(!strncasecmp(cmd_line[1], "kbd", strlen(cmd_line[1]))) {
			kbd_init();
		} else {
			puts("Unknow subsystem for init.");
		}
	} else {
		cpu_init(ip_start);
		fb_init();
		kbd_init();
	}
}


/* On lance la machine (jusqu'a une erreur on Ctrl-C)
 */
static void run(void)
{
	if(!is_code_loaded) {
		file_load_source(DEFAULT_LOAD);
	}

	signal(SIGINT, stop_process);
	proceed = 1;
	ruse = start_clock(&num_of_instr);
	timer_init();
	for(;proceed;) {
		cpu_op();
		num_of_instr++;
	}
}


/* Affichage des registre et pile
 */
static void reg(void)
{
	int ip;
	int w = cpu_get_P(&ip);
	printf("P:\t0x%05X\t(instr: %i)\nA:\t0x%05X\nB:\t0x%05X\nC:\t0x%05X\t(", ip, w, cpu_get_A(), cpu_get_B(), mem_get(REG_CTRL));
	bin_mem(mem_get(REG_CTRL));
	printf(")\n\n");
	for(ip = 15; ip >= 0; ip--) {
		printf("S %2i:\t%s", ip, cpu_get_stack(ip));
		if(ip == cpu_get_top())
			printf(" <=> T");
		else printf("\t");
		printf("\tR %2i:\t%s", ip, cpu_get_rstack(ip));
		if(ip == cpu_get_rtop())
			printf(" <=> R");
		puts("");
	}
}



/* Execution pas-a-pas
 */
static void next()
{
	if(!is_code_loaded) {
		file_load_binary(DEFAULT_LOAD);
	}
	if(cmd_line[1]) {
		if(!ischiffre(cmd_line[1]))
			puts("Bad number of instructions.");
		else {
			int i = atoi(cmd_line[1]);
			while(i-- > 0) {
				cpu_op();
				num_of_instr++;
			}
		}
	} else {
			cpu_op();
			num_of_instr++;
			reg();
	}
}



/* affichage de l'aide
 */
static void help(void)
{
	if(cmd_line[1]) {
		if(!strncasecmp(cmd_line[1], "dump", 4)) {
			puts("\tUsage: dump [x [y]]\n\tPrint memory content from x to y.\n\tDefault: x is instructions pointer. y is x+10.");
		} else if(!strncasecmp(cmd_line[1], "set", 3)) {
			puts("\tUsage: set [keyword value]\n\tChange internal variable.\n\tDefaut: print value.");
		} else if(!strncasecmp(cmd_line[1], "save", 4)) {
			printf("\tUsage: save [filename [x y]]\n\tSave memory into binary file from x to y.\n\tDefault: filename is %s.\n\tIf no x and no y, dump ALL memory.\n", DEFAULT_SAVE_NAME);
		} else if(!strncasecmp(cmd_line[1], "status", 6)) {
			puts("\tUsage: status\n\tPrint informations about VM (interrupt, IO...");
		} else if(!strncmp(cmd_line[1], "register", 8)) {
			puts("\tUsage: register\n\tPrint stack and register contents.");
		} else if(!strncasecmp(cmd_line[1], "help", 4)) {
			puts("\tUsage: help [command]\n\tPrint help or help for command.");
		} else if(!strncasecmp(cmd_line[1], "load", 4)) {
			printf("\tUsage: load [addr]\n\tLoad binary file into memory. Starting at addr.\n\tDefault addr is 0x%05X\n", DEFAULT_LOAD);
		} else if(!strncmp(cmd_line[1], "Run", 3)) {
			puts("\tUsage: Run\n\tStart code execution. Don't stop.");
		} else if(!strncasecmp(cmd_line[1], "init", 4)) {
			puts("\tUsage: init [cpu of kbd of fb]\n\tInitialize sub-system. default is all.");
		} else if(!strncasecmp(cmd_line[1], "compil", 6)) {
			printf("\tUsage: compil [addr]\n\tCompile source file. Starting at addr.\n\tDefault addr is 0x%05X\n", DEFAULT_LOAD);
		} else if(!strncasecmp(cmd_line[1], "next", 4)) {
			puts("\tUsage: run [x]\n\tRun x next instructions.\n\tDefault is 1.");
		} else if(!strncasecmp(cmd_line[1], "quit", 4) || !strncasecmp(cmd_line[1], "exit", 4)) {
			puts("\tUsage: quit or exit\n\tQuit program");
		} else
			puts("\tno help");

	} else
	puts(	"help\t\tShow this help\n"
		"set\t\tSet variable\n"
		"init\t\tInit VM\n"
		"compil\t\tCompile source into memory\n"
		"load\t\tLoad binary code into memory\n"
		"save\t\tSave memory content into file\n"
		"status\t\tShow number of interrupt, IO access...\n"
		"reg\t\tShow internal stack and register\n"
		"dump\t\tShow memory dump\n"
		"Run\t\tExecuting code\n"
		"next\t\tNext instruction\n"
		"quit\t\tExit program");
}


/* on parse la ligne de commande
 */
void parse_cmd(char *buf)
{
	char **ap;

	for(ap = cmd_line; (*ap = strsep(&buf, " \t\n")) != NULL;)
		if(**ap != 0)
			if(++ap >= &cmd_line[NUM_OF_ARG])
				break;

	if(!cmd_line[0])
		return;

	if(!strncasecmp(cmd_line[0], "dump", strlen(cmd_line[0]))) {
		dump();
	} else if(!strncasecmp(cmd_line[0], "set", 3)) {
		set();
	} else if(!strncasecmp(cmd_line[0], "save", 4)) {
		save();
	} else if(!strncasecmp(cmd_line[0], "status", strlen(cmd_line[0]))) {
		status();
	} else if(!strncmp(cmd_line[0], "register", strlen(cmd_line[0]))) {
		reg();
	} else if(!strncasecmp(cmd_line[0], "help", strlen(cmd_line[0]))) {
		help();
	} else if(!strncasecmp(cmd_line[0], "load", strlen(cmd_line[0]))) {
		load();
	} else if(!strncmp(cmd_line[0], "Run", strlen(cmd_line[0]))) {
		run();
	} else if(!strncasecmp(cmd_line[0], "init", strlen(cmd_line[0]))) {
		init();
	} else if(!strncasecmp(cmd_line[0], "compil", strlen(cmd_line[0]))) {
		comp();
	} else if(!strncasecmp(cmd_line[0], "next", strlen(cmd_line[0]))) {
		next();
	} else if(!strncasecmp(cmd_line[0], "quit", strlen(cmd_line[0])) || !strncasecmp(cmd_line[0], "exit", strlen(cmd_line[0]))) {
		quit(0);
	} else
		puts("Unknow command. Type 'h' for help.");
}


/*
int term_init(void)
{
	struct termios cons_temp;

	fcntl(STDIN, F_SETOWN, getpid());
	fcntl(STDIN, F_SETFL, fcntl(STDIN, F_GETFL) | O_ASYNC);
	if(tcgetattr(STDIN, &cons_state)) {
		perror("tcgetattr");
		return -1;
	}
	cons_temp = cons_state;
	cfmakeraw(&cons_temp);
	cons_temp.c_iflag &= ~(IGNBRK|BRKINT|PARMRK|ISTRIP|INLCR|IGNCR|ICRNL|IXON);
	cons_temp.c_oflag &= ~OPOST|OCRNL;
	cons_temp.c_lflag &= ~(ECHO|ECHONL|ICANON|ISIG|IEXTEN);
	cons_temp.c_cflag &= ~(CSIZE|PARENB);
	cons_temp.c_cflag |= CS8;

	if(tcsetattr(STDIN, TCSANOW, &cons_temp)) {
		perror("tcsetattr");
		return -1;
	}

	return 0;
}

void read_cons(int s)
{
	char c;
	static char *s = buf;
	read(STDIN, &c, 1);

	printf("%c", c);
	if(c == 'q') { 
		quit(0);
		return;
	}
	if(c == '\n')
		printf("return");

	parse_cmd(buf);
	write(STDOUT, BLINK"OK > "NOBLINK, 14);
}
*/



/* Le programme
 */
int main(int argc, char *argv[])
{
	char *buffer = jalloc(1024);

	if(argc > 2) {
		printf("Usage: %s [file]\n\n", argv[0]);
		return 0;
	}

	mem_init();
	fb_init();
	kbd_init();

	switch(argc) {
		case 2:
			if(file_open(argv[1]))
				quit(0);

			file_load_source(DEFAULT_LOAD);
		case 1:
			break;
		default:
	}
	
	cpu_init(ip_start);
	int_init();
	//term_init();

	signal(SIGINT, quit);
	signal(SIGTERM, quit);
	//signal(SIGIO, read_cons);
	write(STDOUT, BLINK"OK > "NOBLINK, 14);
	bzero(buffer, 1024);
	while(read(STDIN, buffer, 1024) > 0) {
		parse_cmd(buffer);
		write(STDOUT, BLINK"OK > "NOBLINK, 14);
		bzero(buffer, 1024);
	}

	quit(0);

	/* not reached */
	return 0;
}

