/* $Id: video.c,v 1.14 2002/05/14 15:51:26 jym Exp $
 *
 * Gestion affichage
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "misc.h"
#include "kbd.h"
#include "spec.h"
#include "ansi.h"
#include "debug.h"

extern int stat_videor;
extern int stat_videow;

#ifdef HAVE_SDL
#include <SDL.h>
#include <SDL_syswm.h>


static SDL_Surface *screen = NULL;
static SDL_TimerID timer;


static Uint32 fb_update(Uint32 interval, void *para)
{
	if(SDL_Flip(screen))
		puts(SDL_GetError());

	return interval;
}


void fb_close(void)
{
	DENTER;
	if(screen) {
		SDL_RemoveTimer(timer);
		SDL_Quit();
	}
	DLEAVE;
}



int fb_init(void)
{
	DENTER;
	//printf("Frame buffer (624x420x16)");

	if(SDL_Init(SDL_INIT_VIDEO|SDL_INIT_TIMER|SDL_INIT_EVENTTHREAD) < 0) {
		//puts(PRINT_FAILED);
		DLEAVE;
		return -1;
	}

	if(!SDL_VideoModeOK(624, 420, 16, SDL_HWSURFACE|SDL_DOUBLEBUF|SDL_NOFRAME|SDL_HWACCEL)) {
		//puts(PRINT_FAILED);
	} else {
		screen = SDL_SetVideoMode(624, 420, 16, SDL_HWSURFACE|SDL_DOUBLEBUF|SDL_NOFRAME|SDL_HWACCEL);

		{
			SDL_SysWMinfo info;
			SDL_VERSION(&info.version);
			if ( SDL_GetWMInfo(&info) != 0 ) {
				int x, y;
				int w, h;
				if ( info.subsystem == SDL_SYSWM_X11 ) {
					info.info.x11.lock_func();
					w = DisplayWidth(info.info.x11.display,
					DefaultScreen(info.info.x11.display));
					h = DisplayHeight(info.info.x11.display,
					DefaultScreen(info.info.x11.display));
					x = (w - screen->w)/2;
					y = (h - screen->h)/2;
					XMoveWindow(info.info.x11.display, info.info.x11.wmwindow, x, y);
					info.info.x11.unlock_func();
				}
			}
		}
		//puts(PRINT_OK);
	}

	/* Ce qui fait du 33Hz. Le mieux serait 14 => 71 Hz */
	timer = SDL_AddTimer(30, fb_update, NULL);

	DLEAVE;
	return 0;
}


/* optimisation: mes variables ne servent a rien
 */
int fb_set(int addr, int value)
{
	Uint16 *bufp;
	int x;
	int y;

	DENTER;
	stat_videow++;

	if(!screen) {
		if(DEBUG_VIDEO)
			printf("Write FB at 0x%05X with 0x%05X\n", addr, value);
		DLEAVE;
		return 0;
	}

	x = (addr)%screen->w;
	y = (addr)/screen->w;

	if(x >= 624 || y >= 420) return 0;

	if(DEBUG_VIDEO)
		printf("Write FB %03i,%03i\tColor: 0x%08X\n", x, y, value);

	bufp = (Uint16 *)screen->pixels + y*screen->pitch/2 + x;
	*bufp = value;
	DLEAVE;

	return 0;
}


/* on peut virer les variables la aussi
 */
int fb_get(int addr)
{
	Uint16 *pixel;
	int x;
	int y;

	DENTER;
	stat_videor++;

	if(!screen) {
		if(DEBUG_VIDEO)
			printf("Read FB at 0x%05X\n", addr);
		DLEAVE;
		return 0;
	}

	x = (addr)%screen->w;
	y = (addr)/screen->w;

	if(x >= 624 || y >= 420) return 0;

	if(DEBUG_VIDEO)
		printf("Read FB: %03i,%03i\n", x, y); 

	pixel = (Uint16 *)(screen->pixels + y * screen->pitch + x * (screen->format->BytesPerPixel));


	DLEAVE;
	return *pixel;
}

#else

int *fb;

void fb_close(void)
{
	free(fb);
}


int  fb_init(void)
{
	//printf("Frame buffer (no SDL, using memory)");
	if((fb = jalloc(sizeof(int)*320*240)) == NULL) {
		//puts(PRINT_FAILED);
		return -1;
	}
	//puts(PRINT_OK);

	return 0;
}


void fb_set(int addr, int data)
{
	stat_videow++;

	if(DEBUG_VIDEO)
		printf("Write FB at 0x%05X with 0x%05X\n", addr, data);
	if(fb)
		fb[addr] = data;
}


int fb_get(int addr)
{
	stat_videor++;
	
	if(DEBUG_VIDEO)
		printf("Read FB at 0x%05X\n", addr);
	if(fb)
		return fb[addr];
	else
		return 0;
}

#endif

